package sun.nio.cs;

import java.nio.charset.CoderResult;
import java.nio.charset.CodingErrorAction;
import java.io.UnsupportedEncodingException;
import java.nio.charset.IllegalCharsetNameException;
import java.io.IOException;
import java.nio.CharBuffer;
import java.nio.channels.WritableByteChannel;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.Charset;
import java.io.Writer;

public class StreamEncoder extends Writer
{
    private static final int DEFAULT_BYTE_BUFFER_SIZE = 8192;
    private volatile boolean isOpen;
    private Charset cs;
    private CharsetEncoder encoder;
    private ByteBuffer bb;
    private final OutputStream out;
    private WritableByteChannel ch;
    private boolean haveLeftoverChar;
    private char leftoverChar;
    private CharBuffer lcb;
    static final /* synthetic */ boolean $assertionsDisabled;
    
    private void ensureOpen() throws IOException {
        if (!this.isOpen) {
            throw new IOException("Stream closed");
        }
        return;
    }
    
    public static StreamEncoder forOutputStreamWriter(OutputStream outputStream, Object o, String name) throws UnsupportedEncodingException {
        if (name == null) {
            name = Charset.defaultCharset().name();
        }
        try {
            if (Charset.isSupported(name)) {
                return new StreamEncoder(outputStream, o, Charset.forName(name));
            }
        }
        catch (IllegalCharsetNameException ex) {}
        throw new UnsupportedEncodingException(name);
    }
    
    public static StreamEncoder forOutputStreamWriter(OutputStream outputStream, Object o, Charset charset) {
        return new StreamEncoder(outputStream, o, charset);
    }
    
    public static StreamEncoder forOutputStreamWriter(OutputStream outputStream, Object o, CharsetEncoder charsetEncoder) {
        return new StreamEncoder(outputStream, o, charsetEncoder);
    }
    
    public static StreamEncoder forEncoder(WritableByteChannel writableByteChannel, CharsetEncoder charsetEncoder, int n) {
        return new StreamEncoder(writableByteChannel, charsetEncoder, n);
    }
    
    public String getEncoding() {
        if (this.isOpen()) {
            return this.encodingName();
        }
        return null;
    }
    
    public void flushBuffer() throws IOException {
        synchronized (this.lock) {
            if (this.isOpen()) {
                this.implFlushBuffer();
            }
            else {
                throw new IOException("Stream closed");
            }
        }
    }
    
    @Override
    public void write(int n) throws IOException {
        this.write(new char[] { (char)n }, 0, 1);
    }
    
    @Override
    public void write(char[] array, int n, int n2) throws IOException {
        synchronized (this.lock) {
            this.ensureOpen();
            if (n < 0 || n > array.length || n2 < 0 || n + n2 > array.length || n + n2 < 0) {
                throw new IndexOutOfBoundsException();
            }
            if (n2 == 0) {
                return;
            }
            this.implWrite(array, n, n2);
        }
    }
    
    @Override
    public void write(String s, int srcBegin, int n) throws IOException {
        if (n < 0) {
            throw new IndexOutOfBoundsException();
        }
        char[] dst = new char[n];
        s.getChars(srcBegin, srcBegin + n, dst, 0);
        this.write(dst, 0, n);
        return;
    }
    
    @Override
    public void flush() throws IOException {
        synchronized (this.lock) {
            this.ensureOpen();
            this.implFlush();
        }
    }
    
    @Override
    public void close() throws IOException {
        synchronized (this.lock) {
            if (!this.isOpen) {
                return;
            }
            this.implClose();
            this.isOpen = false;
        }
    }
    
    private boolean isOpen() {
        return this.isOpen;
    }
    
    private StreamEncoder(OutputStream outputStream, Object o, Charset charset) {
        this(outputStream, o, charset.newEncoder().onMalformedInput(CodingErrorAction.REPLACE).onUnmappableCharacter(CodingErrorAction.REPLACE));
    }
    
    private StreamEncoder(OutputStream out, Object lock, CharsetEncoder encoder) {
        super(lock);
        this.isOpen = true;
        this.haveLeftoverChar = false;
        this.lcb = null;
        this.out = out;
        this.ch = null;
        this.cs = encoder.charset();
        this.encoder = encoder;
        if (this.ch == null) {
            this.bb = ByteBuffer.allocate(8192);
        }
    }
    
    private StreamEncoder(WritableByteChannel ch, CharsetEncoder encoder, int n) {
        this.isOpen = true;
        this.haveLeftoverChar = false;
        this.lcb = null;
        this.out = null;
        this.ch = ch;
        this.cs = encoder.charset();
        this.encoder = encoder;
        this.bb = ByteBuffer.allocate((n < 0) ? 8192 : n);
    }
    
    private void writeBytes() throws IOException {
        this.bb.flip();
        int limit = this.bb.limit();
        int position = this.bb.position();
        assert position <= limit;
        int n = (position <= limit) ? (limit - position) : 0;
        if (n > 0) {
            if (this.ch != null) {
                if (this.ch.write(this.bb) != n && !StreamEncoder.$assertionsDisabled) {
                    throw new AssertionError(n);
                }
            }
            else {
                this.out.write(this.bb.array(), this.bb.arrayOffset() + position, n);
            }
        }
        this.bb.clear();
        return;
    }
    
    private void flushLeftoverChar(CharBuffer charBuffer, boolean endOfInput) throws IOException {
        if (!this.haveLeftoverChar && !endOfInput) {
            return;
        }
        if (this.lcb == null) {
            this.lcb = CharBuffer.allocate(2);
        }
        else {
            this.lcb.clear();
        }
        if (this.haveLeftoverChar) {
            this.lcb.put(this.leftoverChar);
        }
        if (charBuffer != null && charBuffer.hasRemaining()) {
            this.lcb.put(charBuffer.get());
        }
        this.lcb.flip();
        while (this.lcb.hasRemaining() || endOfInput) {
            CoderResult encode = this.encoder.encode(this.lcb, this.bb, endOfInput);
            if (encode.isUnderflow()) {
                if (this.lcb.hasRemaining()) {
                    this.leftoverChar = this.lcb.get();
                    if (charBuffer != null && charBuffer.hasRemaining()) {
                        this.flushLeftoverChar(charBuffer, endOfInput);
                    }
                    return;
                }
                else {
                    break;
                }
            }
            else if (encode.isOverflow()) {
                assert this.bb.position() > 0;
                this.writeBytes();
            }
            else {
                encode.throwException();
            }
        }
        this.haveLeftoverChar = false;
        return;
    }
    
    void implWrite(char[] array, int offset, int length) throws IOException {
        CharBuffer wrap = CharBuffer.wrap(array, offset, length);
        if (this.haveLeftoverChar) {
            this.flushLeftoverChar(wrap, false);
        }
        while (wrap.hasRemaining()) {
            CoderResult encode = this.encoder.encode(wrap, this.bb, false);
            if (encode.isUnderflow()) {
                assert wrap.remaining() <= 1 : wrap.remaining();
                if (wrap.remaining() == 1) {
                    this.haveLeftoverChar = true;
                    this.leftoverChar = wrap.get();
                    break;
                }
                break;
            }
            else if (encode.isOverflow()) {
                assert this.bb.position() > 0;
                this.writeBytes();
            }
            else {
                encode.throwException();
            }
        }
    }
    
    void implFlushBuffer() throws IOException {
        if (this.bb.position() > 0) {
            this.writeBytes();
        }
    }
    
    void implFlush() throws IOException {
        this.implFlushBuffer();
        if (this.out != null) {
            this.out.flush();
        }
    }
    
    void implClose() throws IOException {
        this.flushLeftoverChar(null, true);
        while (true) {
            try {
                CoderResult flush = this.encoder.flush(this.bb);
                if (flush.isUnderflow()) {
                    if (this.bb.position() > 0) {
                        this.writeBytes();
                    }
                    if (this.ch != null) {
                        this.ch.close();
                    }
                    else {
                        this.out.close();
                    }
                    break;
                }
                else if (flush.isOverflow()) {
                    assert this.bb.position() > 0;
                    this.writeBytes();
                    continue;
                }
                else {
                    flush.throwException();
                    continue;
                }
            }
            catch (IOException ex) {
                this.encoder.reset();
                throw ex;
            }
        }
    }
    
    String encodingName() {
        return (this.cs instanceof HistoricallyNamedCharset) ? ((HistoricallyNamedCharset)this.cs).historicalName() : this.cs.name();
    }
}
